unit MRHandle;

interface

uses
  Classes, ComCtrls, Controls, DBGrids, Graphics, Grids, SysUtils, TypInfo,
  Windows, Forms, StdCtrls, DBCtrls, FileCtrl, OutLine, Tabs, Buttons, TabNotBk,
  ExtCtrls, DsgnIntf;

type
  TMRControlHandler = class(TObject)
  private
    FControl: TControl;
    FHandlers: TList;

    CLeft, CTop, CWidth, CHeight: Single;
    LastPos: TRect;
    LastParentWidth, LastParentHeight: Integer;
  public
    constructor Create( Control: TControl ); virtual;
    destructor Destroy; override;
    procedure BeforeResize( ParentWidth, ParentHeight: Integer); virtual;
    procedure AfterResize; virtual;
    function GetWarnings: string;

    property Control: TControl read FControl;
    property Handlers: TList read FHandlers;
  end;

   TWarningListProperty = class(TPropertyEditor)
      function GetAttributes: TPropertyAttributes; override;
      procedure Edit; override;
      function GetValue: string; override;
   end;

  TWarningListForm = class(TForm)
    Grid: TStringGrid;
    procedure FormActivate(Sender: TObject);
  private
  public
    Value: TStringList;
  end;


  TMRLSection = class;
  TMRLSections = class;
  TMRLValue = class;
  TMRLValues = class;

  TMRPropertyHandler = class(TObject)
  private
    FControl: TControl;
    procedure Change;       //BUGBUG  Just for debugging

  public
    Invalid: Boolean;

    constructor Create( Control: TControl ); virtual;
    procedure BeforeResize; virtual;
    procedure AfterResize; virtual;
    function GetWarning: string; virtual;

    property Control: TControl read FControl;
  end;

  //***

  TMRFont = class(TMRPropertyHandler)
  private
    FFont: TFont;
    FName: string;
    LastSize: Integer;
    CxSize, CySize: Single;
  public
    constructor Createhandler( Control: TControl; PropertyName: string ); virtual;
    procedure BeforeResize; override;
    procedure BeforeFontResize( CWidth, CHeight: Integer);
    procedure AfterResize; override;
    function GetWarning: string; override;

    property Font: TFont read FFont write FFont;
  end;

  //***

  TMRTab = class(TMRPropertyHandler)
  private
    PInfoWidth, PInfoHeight: PPropInfo;
    LastWidth, LastHeight: Smallint;
    CWidth, Cheight: Single;
  public
    constructor Create( Control: TControl ); override;
    procedure BeforeResize; override;
    procedure AfterResize; override;
  end;

  //***

  TMRScrollBox = class(TMRPropertyHandler)
  private
    FHorz, FVert: TControlScrollBar;
    LastHorz, LastVert: Integer;
    CHRange, CVRange: Single;
  public
    constructor Create( Control: TControl ); override;
    procedure BeforeResize; override;
    procedure AfterResize; override;
  end;

  //***

  TMRStatusBar = class(TMRPropertyHandler)
  private
    FPanels: TStatusPanels;
    FData: TMRLValues;
  public
    constructor Create( Control: TControl ); override;
    destructor Destroy; override;
    procedure BeforeResize; override;
    procedure AfterResize; override;
  end;

  //***

  TMRHeaderControl = class(TMRPropertyHandler)
  private
    FSections: THeaderSections;
    FData: TMRLSections;
  public
    constructor Create( Control: TControl ); override;
    destructor Destroy; override;
    procedure BeforeResize; override;
    procedure AfterResize; override;
  end;

  //***

  TMRDrawGrid = class(TMRPropertyHandler)
  private
    PInfoWidth, PInfoHeight, PInfoLineWidth: PPropInfo;
    LastWidth, LastHeight, LastLineWidth: Integer;
    CWidth, CHeight, CLineWidth: Single;
    FCols: TMRLValues;
    FRows: TMRLValues;
  public
    constructor Create( Control: TControl ); override;
    destructor Destroy; override;
    procedure BeforeResize; override;
    procedure AfterResize; override;
  end;

  //***

  TMRListView = class(TMRPropertyHandler)
  private
    FColumns: TListColumns;
    FData: TMRLValues;
  public
    constructor Create( Control: TControl ); override;
    destructor Destroy; override;
    procedure BeforeResize; override;
    procedure AfterResize; override;
  end;

  //***

  TMRDBGrid = class(TMRPropertyHandler)
  private
    FColumns: TDBGridColumns;
    FData: TMRLValues;
    FFontHandler: TMRFont;
  public
    constructor Create( Control: TControl ); override;
    destructor Destroy; override;
    procedure BeforeResize; override;
    procedure AfterResize; override;
  end;

  //***

  TMRTabSet = class(TMRPropertyHandler)
  private
    PInfoHeight, PInfoStart, PInfoEnd: PPropInfo;
    LastHeight, LastStart, LastEnd: Integer;
    Cheight, CStart, CEnd: Single;
  public
    constructor Create( Control: TControl ); override;
    procedure BeforeResize; override;
    procedure AfterResize; override;
  end;

  //***

  TMRTreeView = class(TMRPropertyHandler)
  private
    PInfoIndent: PPropInfo;
    LastIndent: Integer;
    CIndent: Single;
  public
    constructor Create( Control: TControl ); override;
    procedure BeforeResize; override;
    procedure AfterResize; override;
  end;

  //***

  TMRLSection = class(TCollectionItem)
  private
    FLastWidth, FLastMinWidth, FLastMaxWidth: Integer;
    CWidth, CMinWidth, CMaxWidth: Single;
  published
    property LastMaxWidth: Integer read FLastMaxWidth write FLastMaxWidth;
    property LastMinWidth: Integer read FLastMinWidth write FLastMinWidth;
    property LastWidth: Integer read FLastWidth write FLastWidth;
  end;

  TMRLSections = class(TCollection)
  private
    function GetItem(Index: Integer): TMRLSection;
    procedure SetItem(Index: Integer; Value: TMRLSection);
  public
    function Add: TMRLSection;
    property Items[Index: Integer]: TMRLSection read GetItem write SetItem; default;
  end;

  //***

  TMRItemHeight = class(TMRPropertyHandler)
  private
    PInfoProperty: PPropInfo;
    LastValue: Integer;
    CValue: Single;
  public
    constructor Create( Control: TControl ); override;
    procedure BeforeResize; override;
    procedure AfterResize; override;
  end;

  //***

  TMRMarginSpacing = class(TMRPropertyHandler)
  private
    PInfoMargin, PInfoSpacing: PPropInfo;
    LastMargin, LastSpacing: Integer;
    CMargin, CSpacing: Single;
  public
    constructor Create( Control: TControl ); override;
    procedure BeforeResize; override;
    procedure AfterResize; override;
  end;

  //***

  TMRStretch = class(TMRPropertyHandler)
  public
    function GetWarning: string; override;
  end;

  //***

  TMRAutoSize = class(TMRPropertyHandler)
  public
    function GetWarning: string; override;
  end;

  //***

  TMRLValue = class(TCollectionItem)
  private
    FLastValue: Integer;
    FCValue: Single;
  published
    property LastValue: Integer read FLastValue write FLastValue;
    property CValue: Single read FCValue write FCValue;
  end;

  TMRLValues = class(TCollection)
  private
    function GetItem(Index: Integer): TMRLValue;
    procedure SetItem(Index: Integer; Value: TMRLValue);
  public
    function Add: TMRLValue;
    property Items[Index: Integer]: TMRLValue read GetItem write SetItem; default;
  end;


  procedure AddHandlers( Control: TControl; List: TList );
  function GetPropertyInfo( ClassInfo: Pointer; PropName: String ): Pointer;

var
  WarningListForm: TWarningListForm;
  Class_TCustomComboBox, Class_TDBLookupComboBox: TClass;


implementation

{$R *.DFM}


uses Resizer;



procedure AddHandlers( Control: TControl; List: TList );
var
  pInfo: PPropInfo;
begin
  pInfo := GetPropInfo( Control.ClassInfo, 'Font' );
  if assigned(pInfo) then
    List.Add(TMRFont.CreateHandler( Control, 'Font' ));

  if ((Control is TComboBox) or (Control is TDBComboBox) or
    (Control is TDBListBox) or (Control is TDirectoryListBox) or
    (Control is TFileListBox) or (Control is TOutline) or
    (Control is TListBox)) then
    List.Add(TMRItemHeight.Create( Control ));

  if ((Control is TPageControl) or (Control is TTabControl)) then
    List.Add(TMRTab.Create( Control ));

  if (Control is TScrollBox) then
    List.Add(TMRScrollBox.Create( Control ));

  if (Control is TStatusBar) then
    List.Add(TMRStatusBar.Create( Control ));

  if (Control is THeaderControl) then
    List.Add(TMRHeaderControl.Create( Control ));

  if (Control is TDrawGrid) then
    List.Add(TMRDrawGrid.Create( Control ));

  if (Control is TListView) then
    List.Add(TMRListView.Create( Control ));

  if (Control is TDBGrid) then
    List.Add(TMRDBGrid.Create( Control ));

  if (Control is TTabSet) then
    List.Add(TMRTabSet.Create( Control ));

  if (Control is TTreeView) then
    List.Add(TMRTreeView.Create( Control ));

  if ((Control is TBitBtn) or (Control is TSpeedButton)) then
    List.Add(TMRMarginSpacing.Create( Control ));

  if (Control is TTabbedNotebook) then
    List.Add(TMRFont.CreateHandler( Control, 'TabFont' ));

  if ((Control is TImage) or (Control is TDBImage)) then
    List.Add(TMRStretch.Create( Control ));

  if ((Control is TDBText) or (Control is TLabel)) then
    List.Add(TMRAutoSize.Create( Control ));
end;


//********
//********

constructor TMRControlHandler.Create( Control: TControl );
begin
  CLeft := INVALID_CONTROLCOEFICIENT;

  FControl := Control;
  FHandlers := TList.Create;
end;


destructor TMRControlHandler.Destroy;
begin
  while (FHandlers.Count > 0) do
  begin
    TMRPropertyHandler(FHandlers.Items[0]).Destroy;
    FHandlers.Delete(0);
  end;

  FHandlers.Free;

  inherited;
end;


procedure TMRControlHandler.BeforeResize( ParentWidth, ParentHeight: Integer );
var
  Cnt: Integer;
  NPos: TRect;
  cl, ct, cw, ch, pw, ph: Single;
begin
  // This line is due to the PageControl 'bug'
  // get TabSheet dimensions from the active sheet
  if ((FControl is TTabSheet) and (FControl.Parent is TPageControl)) then
    FControl.BoundsRect := TPageControl(FControl.Parent).ActivePage.BoundsRect;

  // Get control position and dimension
  NPos := FControl.BoundsRect;

  // verify if any change occurred in the control size or position
  // relative to the parent
  if ((LastParentWidth <> ParentWidth) or (LastParentHeight <> ParentHeight) or
    (LastPos.Left <> NPos.Left) or (LastPos.Right <> NPos.Right) or
    (LastPos.Top <> NPos.Top) or (LastPos.Bottom <> NPos.Bottom)) then
    if ((ParentWidth = 0) or (ParentHeight = 0)) then
    begin
      CLeft := INVALID_CONTROLCOEFICIENT;
      LastParentWidth := INVALID_CONTROLSIZE;
    end
    else
    begin
      // calculate position coeficients
      cl := FControl.Left;
      ct := FControl.Top;
      cw := FControl.Width;
      ch := FControl.Height;
      pw := ParentWidth;
      ph := ParentHeight;
      CLeft := cl / pw;
      CTop := ct / ph;
      CWidth := cw / pw;
      CHeight := ch / ph;
    end;

  // Treat other properties
  for Cnt := 0 to FHandlers.Count-1 do
    TMRPropertyHandler(FHandlers.Items[Cnt]).BeforeResize;
end;

procedure TMRControlHandler.AfterResize;
var
  Cnt: Integer;
  R: TRect;
  pInfo: PPropInfo;
  CFont: TFont;
begin
  if (CLeft = INVALID_CONTROLCOEFICIENT) then
    exit;

  // patch for the ComboBox bug, the vertical size is upper limited by the
  // font height
  if ((FControl is Class_TCustomComboBox) or
    (FControl is Class_TDBLookupComboBox)) then
  begin
    pInfo := GetPropInfo( Control.ClassInfo, 'Font' );
    if (pInfo <>  nil) then
    begin
      CFont := TFont(GetOrdProp(Control, pInfo));
      CFont.Size := 1000;
    end;
  end;

  // Get control parent's dimensions
  LastParentWidth := FControl.Parent.ClientWidth;
  LastParentHeight := FControl.Parent.ClientHeight;

  // control new dimension
  R := Bounds(
    Round(CLeft*LastParentWidth),
    Round(CTop*LastParentHeight),
    Round(CWidth*LastParentWidth),
    Round(CHeight*LastParentHeight) );

  // This condition is due to the PageControl 'bug'
  // set TabSheet dimensions through the active sheet
  if ((FControl is TTabSheet) and (FControl.Parent is TPageControl)) then
  begin
    TPageControl(FControl.Parent).ActivePage.BoundsRect := R;
    FControl.BoundsRect := TPageControl(FControl.Parent).ActivePage.BoundsRect;
  end
  else
    // set control dimension
    FControl.BoundsRect := R;

  // Treat other control properties
  for Cnt := 0 to FHandlers.Count-1 do
    TMRPropertyHandler(FHandlers.Items[Cnt]).AfterResize;

  // Set dimensions through the active sheet
  if ((FControl is TTabSheet) and (FControl.Parent is TPageControl)) then
    FControl.BoundsRect := TPageControl(FControl.Parent).ActivePage.BoundsRect;

  // save current position
  LastPos := FControl.BoundsRect;
end;


function TMRControlHandler.GetWarnings: string;
var
  Cnt: Integer;
  st: string;
begin
  Result := '';

  for Cnt := 0 to FHandlers.Count-1 do
  begin
    st := TMRPropertyHandler(FHandlers.Items[Cnt]).GetWarning;
    Result := Result + st;
  end;
end;


{****** Property Editor *****}
function TWarningListProperty.GetAttributes: TPropertyAttributes;
begin
   Result := inherited GetAttributes + [paDialog, paReadOnly] -
    [paSubProperties, paMultiSelect];
end;
                         

procedure TWarningListProperty.Edit;
var
   Dialog: TWarningListForm;
Begin
   Dialog := TWarningListForm.Create( nil );
   try
      Dialog.Value := TStringList(GetOrdValue);
      Dialog.ShowModal;
   finally
      Dialog.Free;
   end;
end;


function TWarningListProperty.GetValue: string;
begin
   Result := 'TWarningList';
end;


{****** Warning List Form *****}
procedure TWarningListForm.FormActivate(Sender: TObject);
var
  Cnt: Integer;
begin
  if (Value.Count > 0) then
    Grid.RowCount := Value.Count+1
  else
  begin
    Grid.RowCount := 2;
    Grid.Cells[0,1] := '';
    Grid.Cells[1,1] := '';
  end;

  Grid.Cells[0,0] := 'Control Name';
  Grid.Cells[1,0] := 'Warning Message';

  for Cnt := 0 to Value.Count-1 do
  begin
    Grid.Cells[0, Cnt+1] := TControl(Value.Objects[Cnt]).Name;
    Grid.Cells[1, Cnt+1] := Value.Strings[Cnt];
  end;
end;


//**************************************
//**************************************
//**************************************

constructor TMRPropertyHandler.Create( Control: TControl );
begin
  FControl := Control;
  Invalid := True;
end;

procedure TMRPropertyHandler.BeforeResize;
begin
end;

procedure TMRPropertyHandler.AfterResize;
begin
end;

function TMRPropertyHandler.GetWarning: string;
begin
  Result := '';
end;

procedure TMRPropertyHandler.Change;
begin
//  MessageBeep(1);
end;

//**************************************
//**************************************
//**************************************
// ******* Font property handler *******

constructor TMRFont.CreateHandler( Control: TControl; PropertyName: string );
var
  pInfo: PPropInfo;
begin
  Create( Control );
  FName := PropertyName;

  // get TFont instance
  pInfo := GetPropertyInfo( Control.ClassInfo, FName );
  FFont := TFont(GetOrdProp(Control, pInfo));
end;


procedure TMRFont.BeforeResize;
var
  fs: Single;
  W, H: Single;
begin
  // verifies if the font size changed
  fs := FFont.Size;
  if ((LastSize <> Trunc(fs)) or (Invalid = True)) then
  begin
    Change; // DEBUG

    W := FControl.ClientWidth;
    H := FControl.ClientHeight;
    if ((W = 0) or (H = 0)) then
    begin
      Invalid := True;
      exit;
    end;

    CxSize := fs / W;
    CySize := fs / H;
  end;

  Invalid := False;
end;

procedure TMRFont.BeforeFontResize( CWidth, CHeight: Integer);
// Aditional function just for handling the MultiResPanel font
// which simple replaces the current dimensions with the given
var
  fs: Single;
  W, H: Single;
begin
  // verifies if the font size changed
  fs := FFont.Size;
  if ((LastSize <> Trunc(fs)) or (Invalid = True)) then
  begin
    Change; // DEBUG

    W := CWidth;
    H := CHeight;
    if ((W = 0) or (H = 0)) then
    begin
      Invalid := True;
      exit;
    end;

    CxSize := fs / W;
    CySize := fs / H;
  end;

  Invalid := False;
end;


procedure TMRFont.AfterResize;
var
  SizeX, SizeY: Single;
  H, W: Single;
begin
  if (Invalid = True) then
    exit;

  W := FControl.ClientWidth;
  H := FControl.ClientHeight;
  SizeX := CxSize * W;
  SizeY := CySize * H;

  if (SizeX < SizeY) then
    FFont.Size := Trunc(SizeX)
  else
    FFont.Size := Trunc(SizeY);

  if (FFont.Size = 0) then
    FFont.Size := 1;

  LastSize := FFont.Size;

  if ((FControl is TTabSheet) and (FControl.Parent is TPageControl)) then
    FControl.BoundsRect := TPageControl(FControl.Parent).ActivePage.BoundsRect;
end;


function TMRFont.GetWarning: string;
var
  tm: TTextMetric;
  OldF, hDC: THandle;
begin
  hDC := GetDC( 0 );
  try
    OldF := SelectObject( hDC, FFont.Handle );
    GetTextMetrics( hDC, tm );
    SelectObject( hDC, OldF );
  finally
    ReleaseDC( 0, hDC );
  end;

  if ((tm.tmPitchAndFamily and TMPF_VECTOR) = 0) then
    Result := 'Not a vector font. '
  else
    Result := '';
end;


//**************************************
//**************************************
//**************************************
// ******* Tab property handler ********

constructor TMRTab.Create( Control: TControl );
begin
  inherited;

  PInfoWidth := GetPropertyInfo( Control.ClassInfo, 'TabWidth' );
  PInfoHeight := GetPropertyInfo( Control.ClassInfo, 'TabHeight' );
end;

procedure TMRTab.BeforeResize;
var
  tw, th: Smallint;
  W, H: Single;
begin
  tw := GetOrdProp( FControl, PInfoWidth );
  th := GetOrdProp( FControl, PInfoHeight );

  // verifies if the tab dimensions changed
  if ((LastWidth <> tw) or (LastHeight <> th) or (Invalid = True)) then
  begin
    Change; // DEBUG

    W := FControl.ClientWidth;
    H := FControl.ClientHeight;

    if ((W = 0) or (H = 0)) then
    begin
      Invalid := True;
      exit;
    end;

    CWidth := tw;
    CWidth := CWidth / W;

    CHeight := th;
    CHeight := CHeight / H;
    Invalid := False;
  end;
end;

procedure TMRTab.AfterResize;
var
  tw, th: Smallint;
begin
  if (Invalid = True) then
    exit;

  tw := GetOrdProp( FControl, PInfoWidth );
  th := GetOrdProp( FControl, PInfoHeight );

  if (tw <> 0) then
  begin
    tw := Trunc(CWidth * FControl.ClientWidth);
    if (tw = 0) then
      tw := 1;

    SetOrdProp( FControl, PInfoWidth, tw );
  end;

  if (th <> 0) then
  begin
    th := Trunc(CHeight * FControl.ClientHeight);
    if (th = 0) then
      th := 1;

    SetOrdProp( FControl, PInfoHeight, th );
  end;

  LastWidth := GetOrdProp( FControl, PInfoWidth );
  LastHeight := GetOrdProp( FControl, PInfoHeight );
end;


//**************************************
//**************************************
//**************************************
// ******* ScrollBox property handler ********

constructor TMRScrollBox.Create( Control: TControl );
var
  pInfo: PPropInfo;
begin
  inherited;

  pInfo := GetPropertyInfo( Control.ClassInfo, 'HorzScrollBar' );
  FHorz := TControlScrollBar(GetOrdProp(Control, pInfo));
  pInfo := GetPropertyInfo( Control.ClassInfo, 'VertScrollBar' );
  FVert := TControlScrollBar(GetOrdProp(Control, pInfo));
end;

procedure TMRScrollBox.BeforeResize;
var
  Hr, Vr: Integer;
  W, H: Single;
begin
  // reset position to avoid plenty of problems
  // must be revised this part
  FHorz.Position := 0;
  FVert.Position := 0;

  Hr := FHorz.Range;
  Vr := FVert.Range;

  // verifies if the tab dimensions changed
  if ((LastHorz <> Hr) or (LastVert <> Vr) or (Invalid = True)) then
  begin
    Change; // DEBUG

    W := FControl.ClientWidth;
    H := FControl.ClientHeight;

    if ((W = 0) or (H = 0)) then
    begin
      Invalid := True;
      exit;
    end;

    CHRange := Hr;
    CHRange := CHRange / W;

    CVRange := Vr;
    CVRange := CVRange / H;
    Invalid := False;
  end;
end;

procedure TMRScrollBox.AfterResize;
begin
  if (Invalid = True) then
    exit;

  FHorz.Range := Trunc(CHRange * FControl.ClientWidth);
  FVert.Range := Trunc(CVRange * FControl.ClientHeight);

  LastHorz := FHorz.Range;
  LastVert := FVert.Range;
end;


//**************************************
//**************************************
//**************************************
// ** Status Panels property handler ***

constructor TMRStatusBar.Create( Control: TControl );
var
  pInfo: PPropInfo;
begin
  inherited;

  FData := TMRLValues.Create( TMRLValue );
  pInfo := GetPropertyInfo( Control.ClassInfo, 'Panels' );
  FPanels := TStatusPanels(GetOrdProp(Control, pInfo));
end;


destructor TMRStatusBar.Destroy;
begin
  FData.Free;

  inherited;
end;

procedure TMRStatusBar.BeforeResize;
var
  W, WC: Single;
  Cnt: Integer;
begin
  // verifies if the panels changed
  if (FPanels.Count <> FData.Count) then
    Invalid := True;

  if (Invalid = False) then
    for Cnt := 0 to FPanels.Count-1 do
      if (FPanels.Items[Cnt].Width <> FData.Items[Cnt].LastValue) then
      begin
        Invalid := True;
        break;
      end;

  if (Invalid = False) then
    exit;

  Change; // DEBUG

  WC := FControl.ClientWidth;
  if (WC = 0) then
  begin
    Invalid := True;
    exit;
  end;

  FData.Clear;
  for Cnt := 0 to FPanels.Count-1 do
  begin
    W := FPanels.Items[Cnt].Width;
    FData.Add;  // BUGBUG it's assumed the item is added at end of array
    FData.Items[Cnt].CValue := W / WC;
  end;

  Invalid := False;
end;


procedure TMRStatusBar.AfterResize;
var
  Cnt: Integer;
  W: Single;
begin
  if (Invalid = True) then
    exit;

  W := FControl.ClientWidth;

  for Cnt := 0 to FData.Count-1 do
  begin
    FPanels.Items[Cnt].Width := Trunc(FData.Items[Cnt].CValue * W);
    FData.Items[Cnt].LastValue := FPanels.Items[Cnt].Width;
  end;
end;


//**************************************
//**************************************
//**************************************
// ** HeaderSections property handler **

constructor TMRHeaderControl.Create( Control: TControl );
var
  pInfo: PPropInfo;
begin
  inherited;

  FData := TMRLSections.Create( TMRLSection );
  pInfo := GetPropertyInfo( Control.ClassInfo, 'Sections' );
  FSections := THeaderSections(GetOrdProp(Control, pInfo));
end;


destructor TMRHeaderControl.Destroy;
begin
  FData.Free;

  inherited;
end;

procedure TMRHeaderControl.BeforeResize;
var
  Changed: Byte;
  W, WC: Single;
  Cnt: Integer;
begin
  // verifies if the sections changed
  Changed := $00;
  // verifies if number of sections differ
  if ((FSections.Count <> FData.Count) or (Invalid = True)) then
    Changed := $FF
  else
  begin
    // verifies if sections width differ, which can happen by user action
    // if so recalc at least this coeficient
    for Cnt := 0 to FSections.Count-1 do
      if (FSections.Items[Cnt].Width <> FData.Items[Cnt].LastWidth) then
      begin
        Changed := $0F;
        break;
      end;

    // verifies if sections min/max widths differ
    // if so recalc all coeficients
    for Cnt := 0 to FSections.Count-1 do
      if ((FSections.Items[Cnt].MinWidth <> FData.Items[Cnt].LastMinWidth) or
        (FSections.Items[Cnt].MaxWidth <> FData.Items[Cnt].LastMaxWidth)) then
      begin
        Changed := $FF;
        break;
      end;
  end;

  if (Changed = $00) then
    exit;

  Change; // DEBUG

  WC := FControl.ClientWidth;
  if (WC = 0) then
  begin
    Invalid := True;
    exit;
  end;

  // calc min/max width coeficients
  if ((Changed and $F0) <> 0) then
  begin
    FData.Clear;
    for Cnt := 0 to FSections.Count-1 do
    begin
      FData.Add;  // BUGBUG it's assumed the item is added at end of array
      W := FSections.Items[Cnt].MinWidth;
      FData.Items[Cnt].CMinWidth := W / WC;
      W := FSections.Items[Cnt].MaxWidth;
      FData.Items[Cnt].CMaxWidth := W / WC;
    end;
  end;

  // calc width coeficient
  if ((Changed and $0F) <> 0) then
    for Cnt := 0 to FSections.Count-1 do
    begin
      W := FSections.Items[Cnt].Width;
      FData.Items[Cnt].CWidth := W / WC;
    end;

  Invalid := False;
end;

procedure TMRHeaderControl.AfterResize;
var
  Cnt: Integer;
  W: Single;
begin
  if (Invalid = True) then
    exit;

  W := FControl.ClientWidth;

  for Cnt := 0 to FData.Count-1 do
  begin
    FSections.Items[Cnt].Width := Trunc(FData.Items[Cnt].CWidth * W);
    FSections.Items[Cnt].MinWidth := Trunc(FData.Items[Cnt].CMinWidth * W);
    FSections.Items[Cnt].MaxWidth := Trunc(FData.Items[Cnt].CMaxWidth * W);

    FData.Items[Cnt].LastWidth := FSections.Items[Cnt].Width;
    FData.Items[Cnt].LastMinWidth := FSections.Items[Cnt].MinWidth;
    FData.Items[Cnt].LastMaxWidth := FSections.Items[Cnt].MaxWidth;
  end;
end;


//**************************************
//**************************************
//**************************************
// ** Col/Row Grid size property handler ***

constructor TMRDrawGrid.Create( Control: TControl );
begin
  inherited;

  FCols := TMRLValues.Create( TMRLValue );
  FRows := TMRLValues.Create( TMRLValue );

  PInfoWidth := GetPropertyInfo( Control.ClassInfo, 'DefaultColWidth' );
  PInfoHeight := GetPropertyInfo( Control.ClassInfo, 'DefaultRowHeight' );
  PInfoLineWidth := GetPropertyInfo( Control.ClassInfo, 'GridLineWidth' );
end;


destructor TMRDrawGrid.Destroy;
begin
  FCols.Free;
  FRows.Free;

  inherited;
end;

procedure TMRDrawGrid.BeforeResize;
var
  tw, th, lw: Integer;
  W, H, WC, HC: Single;
  Cnt: Integer;
begin
  // verifies changes
  tw := GetOrdProp( FControl, PInfoWidth );
  th := GetOrdProp( FControl, PInfoHeight );
  lw := GetOrdProp( FControl, PInfoLineWidth );
  WC := FControl.ClientWidth;
  HC := FControl.ClientHeight;

  if ((LastWidth <> tw) or (LastHeight <> th)) then
    Invalid := True;

  with FControl as TDrawGrid do
  begin
    // verifies if the cells dimensions changed
    if ((FCols.Count <> ColCount) or (FRows.Count <> RowCount)) then
      Invalid := True
    else
    begin
      for Cnt := 0 to FCols.Count-1 do
        if (FCols.Items[Cnt].LastValue <> ColWidths[Cnt]) then
        begin
          Invalid := True;
          break;
        end;

      for Cnt := 0 to FRows.Count-1 do
        if (FRows.Items[Cnt].LastValue <> RowHeights[Cnt]) then
        begin
          Invalid := True;
          break;
        end;
    end;
  end;

  if (Invalid = True) then
  begin
    Change; // DEBUG

    if ((WC = 0) or (HC = 0)) then
      exit;

    CWidth := tw;
    CWidth := CWidth / WC;

    CHeight := th;
    CHeight := CHeight / HC;

    CLineWidth := lw;
    CLineWidth := CLineWidth / WC;

    with FControl as TDrawGrid do
    begin
      FCols.Clear;
      for Cnt := 0 to ColCount-1 do
      begin
        W := ColWidths[Cnt];
        FCols.Add;  // BUGBUG it's assumed the item is added at end of array
        FCols.Items[Cnt].CValue := W / WC;
      end;

      FRows.Clear;
      for Cnt := 0 to RowCount-1 do
      begin
        H := RowHeights[Cnt];
        FRows.Add;  // BUGBUG it's assumed the item is added at end of array
        FRows.Items[Cnt].CValue := H / HC;
      end;
    end;
  end;

  Invalid := False;
end;


procedure TMRDrawGrid.AfterResize;
var
  Cnt: Integer;
  W: Single;
begin
  if (Invalid = True) then
    exit;

  SetOrdProp( FControl, PInfoWidth, Trunc(CWidth * FControl.ClientWidth) );
  SetOrdProp( FControl, PInfoHeight, Trunc(CHeight * FControl.ClientHeight) );
  SetOrdProp( FControl, PInfoLineWidth, Trunc(CLineWidth * FControl.ClientHeight) );

  LastWidth := GetOrdProp( FControl, PInfoWidth );
  LastHeight := GetOrdProp( FControl, PInfoHeight );
  LastLineWidth := GetOrdProp( FControl, PInfoLineWidth );

  with FControl as TDrawGrid do
  begin
    W := FControl.ClientWidth;
    for Cnt := 0 to FCols.Count-1 do
      ColWidths[Cnt] := Trunc(FCols.Items[Cnt].CValue * W);

    W := FControl.ClientHeight;
    for Cnt := 0 to FRows.Count-1 do
      RowHeights[Cnt] := Trunc(FRows.Items[Cnt].CValue * W);

    for Cnt := 0 to FCols.Count-1 do
      FCols.Items[Cnt].LastValue := ColWidths[Cnt];
    for Cnt := 0 to FRows.Count-1 do
      FRows.Items[Cnt].LastValue := RowHeights[Cnt];
  end;
end;


//**************************************
//**************************************
//**************************************
// ** List Columns property handler ***

constructor TMRListView.Create( Control: TControl );
var
  pInfo: PPropInfo;
begin
  inherited;

  FData := TMRLValues.Create( TMRLValue );
  pInfo := GetPropertyInfo( Control.ClassInfo, 'Columns' );
  FColumns := TListColumns(GetOrdProp(Control, pInfo));
end;


destructor TMRListView.Destroy;
begin
  FData.Free;

  inherited;
end;

procedure TMRListView.BeforeResize;
var
  W, WC: Single;
  Cnt: Integer;
begin
  // verifies if the panels changed
  if ((FColumns.Count <> FData.Count) or (Invalid = True)) then
    Invalid := True
  else
  begin
    for Cnt := 0 to FColumns.Count-1 do
      if (FColumns.Items[Cnt].Width <> FData.Items[Cnt].LastValue) then
      begin
        Invalid := True;
        break;
      end;
  end;

  if (Invalid = False) then
    exit;

  Change; // DEBUG

  // certifies that the width isn't 0
  WC := FControl.ClientWidth;
  if (WC = 0) then
    exit;

  FData.Clear;
  for Cnt := 0 to FColumns.Count-1 do
  begin
    W := FColumns.Items[Cnt].Width;
    FData.Add;  // BUGBUG it's assumed the item is added at end of array
    FData.Items[Cnt].CValue := W / WC;
  end;

  Invalid := False;
end;

procedure TMRListView.AfterResize;
var
  Cnt: Integer;
  W: Single;
begin
  if (Invalid = True) then
    exit;

  W := FControl.ClientWidth;

  for Cnt := 0 to FData.Count-1 do
  begin
    FColumns.Items[Cnt].Width := Trunc(FData.Items[Cnt].CValue * W);

    FData.Items[Cnt].LastValue := FColumns.Items[Cnt].Width;
  end;
end;


//**************************************
//**************************************
//**************************************
// ** List Columns property handler ***

constructor TMRDBGrid.Create( Control: TControl );
var
  pInfo: PPropInfo;
begin
  inherited;

  FData := TMRLValues.Create( TMRLValue );
  pInfo := GetPropertyInfo( Control.ClassInfo, 'Columns' );
  FColumns := TDBGridColumns(GetOrdProp(Control, pInfo));

  FFontHandler := TMRFont.CreateHandler( Control, 'TitleFont' );
end;


destructor TMRDBGrid.Destroy;
begin
  FFontHandler.Free;

  FData.Free;

  inherited;
end;

procedure TMRDBGrid.BeforeResize;
var
  W, WC: Single;
  Cnt: Integer;
begin
  FFontHandler.BeforeResize;

  // verifies if the panels changed
  if ((FColumns.Count <> FData.Count) or (Invalid = True)) then
    Invalid := True
  else
  begin
    for Cnt := 0 to FColumns.Count-1 do
      if (FColumns.Items[Cnt].Width <> FData.Items[Cnt].LastValue) then
      begin
        Invalid := True;
        break;
      end;
  end;

  if (Invalid = False) then
    exit;

  Change; // DEBUG

  // certifies that the width isn't 0
  WC := FControl.ClientWidth;
  if (WC = 0) then
    exit;

  FData.Clear;
  for Cnt := 0 to FColumns.Count-1 do
  begin
    W := FColumns.Items[Cnt].Width;
    FData.Add;  // BUGBUG it's assumed the item is added at end of array
    FData.Items[Cnt].CValue := W / WC;
  end;

  Invalid := False;
end;

procedure TMRDBGrid.AfterResize;
var
  Cnt: Integer;
  W: Single;
begin
  FFontHandler.AfterResize;

  if (Invalid = True) then
    exit;

  W := FControl.ClientWidth;

  for Cnt := 0 to FData.Count-1 do
  begin
    FColumns.Items[Cnt].Width := Trunc(FData.Items[Cnt].CValue * W);

    FData.Items[Cnt].LastValue := FColumns.Items[Cnt].Width;
  end;
end;


//**************************************
//**************************************
//**************************************
// ******* TabSet property handler *****

constructor TMRTabSet.Create( Control: TControl );
begin
  inherited;

  PInfoHeight := GetPropertyInfo( Control.ClassInfo, 'TabHeight' );
  PInfoStart := GetPropertyInfo( Control.ClassInfo, 'StartMargin' );
  PInfoEnd := GetPropertyInfo( Control.ClassInfo, 'EndMargin' );
end;

procedure TMRTabSet.BeforeResize;
var
  th, sm, em: Integer;
  W, H: Single;
begin
  // verifies if any property changed
  th := GetOrdProp( FControl, PInfoHeight );
  sm := GetOrdProp( FControl, PInfoStart );
  em := GetOrdProp( FControl, PInfoEnd );

  if ((LastHeight <> th) or (LastStart <> sm) or (LastEnd <> em)
    or (Invalid = True)) then
  begin
    Change; // DEBUG

    H := FControl.ClientHeight;
    W := FControl.ClientWidth;
    if ((H = 0) or (W = 0)) then
    begin
      Invalid := True;
      exit;
    end;

    CHeight := th;
    CHeight := CHeight / H;

    CStart := sm;
    CStart := CStart / W;

    CEnd := em;
    CEnd := CEnd / W;
  end;

  Invalid := False;
end;

procedure TMRTabSet.AfterResize;
begin
  if (Invalid = True) then
    exit;

  SetOrdProp( FControl, PInfoHeight, Trunc(CHeight * FControl.ClientHeight) );
  SetOrdProp( FControl, PInfoStart, Trunc(CStart * FControl.ClientWidth) );
  SetOrdProp( FControl, PInfoEnd, Trunc(CEnd * FControl.ClientWidth) );

  LastHeight := GetOrdProp( FControl, PInfoHeight );
  LastStart := GetOrdProp( FControl, PInfoStart );
  LastEnd := GetOrdProp( FControl, PInfoEnd );
end;


//**************************************
//**************************************
//**************************************
//**** ItemHeight property handler *****

constructor TMRItemHeight.Create( Control: TControl );
begin
  inherited;

  PInfoProperty := GetPropertyInfo( Control.ClassInfo, 'ItemHeight' );
end;

procedure TMRItemHeight.BeforeResize;
var
  th: Integer;
  H: Single;
begin
  // verifies if the property changed
  th := GetOrdProp( FControl, PInfoProperty );

  if ((LastValue <> th) or (Invalid = True)) then
  begin
    Change; // DEBUG

    H := FControl.ClientHeight;
    if (H = 0) then
    begin
      Invalid := True;
      exit;
    end;

    CValue := th;
    CValue := CValue / H;
  end;

  Invalid := False;
end;

                     
procedure TMRItemHeight.AfterResize;
begin
  if (Invalid = True) then
    exit;

  SetOrdProp( FControl, PInfoProperty, Trunc(CValue * FControl.ClientHeight) );
                     
  LastValue := GetOrdProp( FControl, PInfoProperty );
end;


//**************************************
//**************************************
//**************************************
//* Space & Margin properties handler **

constructor TMRMarginSpacing.Create( Control: TControl );
begin
  inherited;

  PInfoMargin := GetPropertyInfo( Control.ClassInfo, 'Margin' );
  PInfoSpacing := GetPropertyInfo( Control.ClassInfo, 'Spacing' );
end;


procedure TMRMarginSpacing.BeforeResize;
var
  Margin, Spacing: Integer;
  W: Single;
begin
  // verifies if the properties changed
  Margin := GetOrdProp( FControl, PInfoMargin );
  Spacing := GetOrdProp( FControl, PInfoSpacing );

  W := FControl.ClientWidth;

  if (Margin = -1) then
  begin
    Invalid := True;
    exit;
  end;

  if ((LastMargin <> Margin) or (LastSpacing <> Spacing)) then
    Invalid := True;

  if (Invalid = True) then
  begin
    Change; // DEBUG

    if (W = 0) then
      exit;

    CMargin := Margin;
    CMargin := CMargin / W;

    CSpacing := Spacing;
    CSpacing := CSpacing / W;
  end;

  Invalid := False;
end;


procedure TMRMarginSpacing.AfterResize;
begin
  if (Invalid = False) then
  begin
    SetOrdProp( FControl, PInfoMargin, Trunc(CMargin * FControl.ClientWidth) );
    LastMargin := GetOrdProp( FControl, PInfoMargin );

    SetOrdProp( FControl, PInfoSpacing, Trunc(CSpacing * FControl.ClientWidth) );
    LastSpacing := GetOrdProp( FControl, PInfoSpacing );
  end;
end;


//**************************************
//**************************************
//**************************************
// ***** TreeView property handler *****

constructor TMRTreeView.Create( Control: TControl );
begin
  inherited;

  PInfoIndent := GetPropertyInfo( Control.ClassInfo, 'Indent' );
end;

procedure TMRTreeView.BeforeResize;
var
  ind: Integer;
  W: Single;
begin
  // verifies if property changed
  ind := GetOrdProp( FControl, PInfoIndent );

  if ((LastIndent <> ind) or (Invalid = True)) then
  begin
    Change; // DEBUG

    W := FControl.ClientWidth;
    if (W = 0) then
    begin
      Invalid := True;
      exit;
    end;

    CIndent := ind;
    CIndent := CIndent / W;
  end;

  Invalid := False;
end;


procedure TMRTreeView.AfterResize;
begin
  if (Invalid = True) then
    exit;

  SetOrdProp( FControl, PInfoIndent, Trunc(CIndent * FControl.ClientWidth) );

  LastIndent := GetOrdProp( FControl, PInfoIndent );
end;



//**************************************
//**************************************
//**************************************
// ******* Stretch property handler *******


function TMRStretch.GetWarning: string;
var
  pInfo: PPropInfo;
  b: LongBool;
begin
  pInfo := GetPropertyInfo( Control.ClassInfo, 'Stretch' );
  b := LongBool(GetOrdProp(Control, pInfo));

  if (b = False) then
    Result := 'Set Stretch to True'
  else
    Result := '';
end;

//**************************************
//**************************************
//**************************************
// ******* AutoSize property handler *******


function TMRAutoSize.GetWarning: string;
var
  pInfo: PPropInfo;
  b: LongBool;
begin
  pInfo := GetPropertyInfo( Control.ClassInfo, 'AutoSize' );
  b := LongBool(GetOrdProp(Control, pInfo));

  if (b = True) then
    Result := 'Set AutoSize to False'
  else
    Result := '';
end;

//**************************************
//**************************************
//**************************************

function TMRLSections.Add: TMRLSection;
begin
  Result := TMRLSection(inherited Add);
end;

function TMRLSections.GetItem(Index: Integer): TMRLSection;
begin
  Result := TMRLSection(inherited GetItem(Index));
end;

procedure TMRLSections.SetItem(Index: Integer; Value: TMRLSection);
begin
  inherited SetItem(Index, Value);
end;


function TMRLValues.Add: TMRLValue;
begin
  Result := TMRLValue(inherited Add);
end;

function TMRLValues.GetItem(Index: Integer): TMRLValue;
begin
  Result := TMRLValue(inherited GetItem(Index));
end;

procedure TMRLValues.SetItem(Index: Integer; Value: TMRLValue);
begin
  inherited SetItem(Index, Value);
end;


function GetPropertyInfo( ClassInfo: Pointer; PropName: String ): Pointer;
begin
  Result := GetPropInfo( ClassInfo, PropName );
  if (Result = nil) then
    raise Exception.Create('No ' + PropName + ' property');
end;

initialization

Class_TCustomComboBox := TCustomComboBox;
Class_TDBLookupComboBox := TDBLookupComboBox;


end.
